/*
 * Segment 6 Intelligente Bewertungsverfahren, SWE 6.1 Stauverlaufsanalyse
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.ibv.anastau.spez;

import java.util.Collection;

import org.junit.Assert;

import com.bitctrl.Constants;

import de.bsvrz.ibv.anastau.spez.SpezifikationIT.TestVariante;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.objekte.Stau;
import de.bsvrz.sys.funclib.bitctrl.modell.verkehr.onlinedaten.OdStauVerlauf;

/**
 * Klasse zur Prüfung der Ergebnisse der Testausführung.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
public class PhasenPruefer {

	/**
	 * Eintrag für die Definition der Daten eines erwarteten Stauobjekts.
	 *
	 * @author BitCtrl Systems GmbH, Uwe Peuker
	 */
	public static class ErgebnisEintrag {

		/** die mit der Prognose ermittelte maximale Länge des Staus. */
		private final Long maxLaenge;

		/** die mit der Prognose ermittelte maximale Verlustzeit. */
		private final Long verlustZeit;

		/** die aktuelle Geschwindigkeit im Stau. */
		private final Long geschwindigkeit;

		/** die aktuelle Länge des Staus. */
		private final double laenge;

		/**
		 * Konstruktor.
		 *
		 * @param laenge
		 *            die aktuelle Staulänge
		 * @param maxLaenge
		 *            die maximale prognostizierte Länge
		 * @param verlustZeit
		 *            die Verlustzeit
		 * @param geschwindigkeit
		 *            die aktuelle Geschwindigkeit
		 */
		public ErgebnisEintrag(final double laenge, final Long maxLaenge, final Long verlustZeit,
				final Long geschwindigkeit) {
			super();
			this.laenge = laenge;
			this.maxLaenge = maxLaenge;
			this.verlustZeit = verlustZeit;
			this.geschwindigkeit = geschwindigkeit;
		}

		/**
		 * liefert die Geschwindigkeit oder <code>null</code>, wenn kein Wert
		 * erwartet wird.
		 *
		 * @return den Wert oder <code>null</code>
		 */
		public Long getGeschwindigkeit() {
			return geschwindigkeit;
		}

		/**
		 * liefert die erwartete Staulänge.
		 *
		 * @return den Wert oder <code>null</code>
		 */
		public double getLaenge() {
			return laenge;
		}

		/**
		 * liefert die maximal erwartete prognostizierte Staulänge.
		 *
		 * @return die Länge oder <code>null</code>, wenn kein Wert erwartet
		 *         wird
		 */
		public Long getMaxLaenge() {
			return maxLaenge;
		}

		/**
		 * liefert die erwartete Verlustzeit oder <code>null</code>, wenn keine
		 * erwartet wird.
		 *
		 * @return die Zeit oder <code>null</code>
		 */
		public Long getVerlustZeit() {
			return verlustZeit;
		}
	}

	/** die Nummer der Phase, die geprüft wird. */
	private final int phase;

	/** die erwarteten Ergebnisse für den Hauptprüfungspfad. */
	private final ErgebnisEintrag[][] ergebnisse = {
			// Phase 0
			{},
			// Phase 1
			{ new ErgebnisEintrag(180.0, 352L, 93L, 14L), new ErgebnisEintrag(2090.0, 2642L, 1554L, 6L) },
			// Phase 2
			{ new ErgebnisEintrag(264.6, 412L, 118L, 13L), new ErgebnisEintrag(2415.5, 3470L, 2379L, 5L) },
			// Phase 3
			{ new ErgebnisEintrag(412.05, 412L, 99L, 15L), new ErgebnisEintrag(2910.95, 3984L, 3230L, 4L) },
			// Phase 4
			{ new ErgebnisEintrag(279.90, 323L, 95L, 12L), new ErgebnisEintrag(3447.35, 4574L, 3977L, 4L) },
			// Phase 5
			{ new ErgebnisEintrag(297.90, 298L, 69L, 16L), new ErgebnisEintrag(4007.90, 5206L, 4881L, 4L) },
			// Phase 6
			{ new ErgebnisEintrag(158.40, 257L, 83L, 11L), new ErgebnisEintrag(4602.20, 5648L, 5213L, 4L) },
			// Phase 7
			{ new ErgebnisEintrag(6124.95, 7045L, 6038L, 4L) },
			// Phase 8
			{ new ErgebnisEintrag(6573.15, 6989L, 2912L, 9L) },
			// Phase 9
			{ new ErgebnisEintrag(1910.0, 2137L, 801L, 10L), new ErgebnisEintrag(874.05, 874L, 256L, 12L) },
			// Phase 10
			{ new ErgebnisEintrag(2045.0, 3319L, 3319L, 4L), new ErgebnisEintrag(556.95, 557L, 154L, 13L) },
			// Phase 11
			{ new ErgebnisEintrag(1910.0, 1955L, 902L, 8L), new ErgebnisEintrag(239.10, 239L, 99L, 9L) } };

	/** die erwarteten Ergebnisse für die Variante A. */
	private final ErgebnisEintrag[][] ergebnisseA = {
			// Phase 0
			{},
			// Phase 1
			{ new ErgebnisEintrag(180.0, null, null, null), new ErgebnisEintrag(180.0, null, null, null),
					new ErgebnisEintrag(1000.0, null, null, null) } };

	/** die erwarteten Ergebnisse für die Variante B. */
	private final ErgebnisEintrag[][] ergebnisseB = {
			// Phase 0
			{},
			// Phase 1
			{ new ErgebnisEintrag(180.0, 352L, 93L, 14L), new ErgebnisEintrag(2090.0, 2642L, 1554L, 6L) },
			// Phase 2
			{ new ErgebnisEintrag(264.6, 412L, 118L, 13L), new ErgebnisEintrag(2415.5, 3470L, 2379L, 5L) },
			// Phase 3
			{ new ErgebnisEintrag(412.05, 412L, 99L, 15L), new ErgebnisEintrag(2910.95, 3984L, 3230L, 4L) },
			// Phase 4
			{ new ErgebnisEintrag(279.90, 323L, 95L, 12L), new ErgebnisEintrag(3447.35, 4574L, 3977L, 4L) },
			// Phase 5
			{ new ErgebnisEintrag(297.90, 298L, 69L, 16L), new ErgebnisEintrag(4007.90, 5206L, 4881L, 4L) },
			// Phase 6
			{ new ErgebnisEintrag(158.40, 257L, 83L, 11L), new ErgebnisEintrag(4602.20, 5648L, 5213L, 4L) },
			// Phase 7
			{ new ErgebnisEintrag(215, null, null, null), new ErgebnisEintrag(4603, null, null, null) }, };

	/** die erwarteten Ergebnisse für die Variante D. */
	private final ErgebnisEintrag[][] ergebnisseD = {
			// Phase 0
			{},
			// Phase 1
			{ new ErgebnisEintrag(180.0, 352L, 93L, 14L), new ErgebnisEintrag(2090.0, 2642L, 1554L, 6L) },
			// Phase 2
			{ new ErgebnisEintrag(264.6, 412L, 118L, 13L), new ErgebnisEintrag(2415.5, 3470L, 2379L, 5L) },
			// Phase 3
			{ new ErgebnisEintrag(412.05, 412L, 99L, 15L), new ErgebnisEintrag(2910.95, 3984L, 3230L, 4L) },
			// Phase 4
			{ new ErgebnisEintrag(279.90, 323L, 95L, 12L), new ErgebnisEintrag(3447.35, 4574L, 3977L, 4L) },
			// Phase 5
			{ new ErgebnisEintrag(297.90, 298L, 69L, 16L), new ErgebnisEintrag(4007.90, 5206L, 4881L, 4L) },
			// Phase 6
			{ new ErgebnisEintrag(158.40, 257L, 83L, 11L), new ErgebnisEintrag(4602.20, 5648L, 5213L, 4L) },
			// Phase 7
			{ new ErgebnisEintrag(6124.95, 7045L, 6038L, 4L) },
			// Phase 8
			{ new ErgebnisEintrag(6573.15, 6989L, 2912L, 9L) },
			// Phase 9
			{ new ErgebnisEintrag(6784.05, null, null, null) } };

	/** die verwendete Testvariante. */
	private final TestVariante testVariante;

	/**
	 * Konstruktor. Die Funktion erzeugt ein Prüfungsobjekt für die übergebene
	 * Phase der Prüfung.
	 *
	 * @param phase
	 *            die Phase
	 * @param variante
	 *            die verwendete Testvariante
	 */
	public PhasenPruefer(final int phase, final TestVariante variante) {
		this.phase = phase;
		this.testVariante = variante;
	}

	/**
	 * prüft, ob die ermittelten Staus den erwarteten Ergebnissen entsprechen.
	 *
	 * @param staus
	 *            die Staus, die während der Phase angelegt wurde
	 */
	public void pruefe(final Collection<Stau> staus) {
		ErgebnisEintrag[] erwartet = null;
		switch (testVariante) {
		case A:
			erwartet = ergebnisseA[phase];
			break;
		case B:
			erwartet = ergebnisseB[phase];
			break;
		case C:
		case D:
			erwartet = ergebnisseD[phase];
			break;
		case MAIN:
		default:
			erwartet = ergebnisse[phase];
			break;
		}

		if (erwartet.length != staus.size()) {
			Assert.fail("In Phase " + phase + " wurden " + erwartet.length + " Stauobjekte erwartet, ermittelt wurden "
					+ staus.size());
		}

		for (final Stau stau : staus) {
			sucheErwartetenStau(stau, erwartet);
		}
	}

	/**
	 * ermittelt, ob der übergebene Stau einem der für die aktuelle Phase
	 * erwarteten entspricht.
	 *
	 * @param stau
	 *            der Stau, der geprüft wird
	 * @param erwartet
	 *            die erwarteten Staudaten
	 */
	private void sucheErwartetenStau(final Stau stau, final ErgebnisEintrag[] erwartet) {

		boolean gefunden = false;
		final double stauLaenge = stau.getSituationsEigenschaften().getDatum().getLaenge();
		final OdStauVerlauf.Daten datum = stau.getOnlineDatensatz(OdStauVerlauf.class)
				.getDatum(OdStauVerlauf.Aspekte.PrognoseNormal.getAspekt());
		if (datum == null) {
			Assert.fail("Für ein Stauobjekt konnten in Phase " + phase + " keine Prognosedaten ermittelt werden");
		} else {

			for (final ErgebnisEintrag eintrag : erwartet) {
				if ((eintrag.getGeschwindigkeit() == null)
						|| (eintrag.getGeschwindigkeit() == datum.getAktuelleGeschwindigkeit())) {
					if ((eintrag.getMaxLaenge() == null) || (eintrag.getMaxLaenge() == datum.getMaxLaenge())) {
						if ((eintrag.getVerlustZeit() == null) || (eintrag.getVerlustZeit() == Math
								.round(((double) datum.getMaxVerlustzeit() / Constants.MILLIS_PER_SECOND)))) {
							if (Math.abs(stauLaenge - eintrag.getLaenge()) <= 1.0) {
								gefunden = true;
								break;
							}
						}
					}
				}
			}

			if (!gefunden) {
				Assert.fail("In Phase " + phase + " wurden die Staudaten: Staulänge: " + stauLaenge + ", MaxLänge: "
						+ datum.getMaxLaenge() + ", Verlustzeit: " + datum.getMaxVerlustzeit() + ", Geschwindigkeit: "
						+ datum.getAktuelleGeschwindigkeit() + " nicht erwartet");
			}
		}
	}
}
